/**
 * Introduction
 *
 * @title ROV - Easy applications from C
 *
 * @copyright Jason Tribbeck / 7th ARM 1994-2001
 *
 * @authors Jason Tribbeck
 *
 * @file list.c
 */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>

#include "kernel.h"

#ifndef __GNUC__
#pragma -s1
#endif

#include "main.h"
#include "canvas.h"
#include "icons.h"
#include "menu.h"
#include "msg.h"
#include "draw.h"
#include "drawprintf.h"

typedef void (* list_print)(void * t, int i, int x, int y, void * h, int wimpflags, int fore, int back);

typedef struct {
  int flags, y;
  void * spra;
  char sprn[12];
  union {
    char * offset;
    char name[12];
  } text;
  void * t;
  void * handle;
  int dispflags;
} list_item;

typedef struct {
  int x, y;
  int items, current;
  int height;
  void * parent;
  list_print p;
  list_item item[1];
} list_tree;

typedef struct {
  int flags;
  union {
    char name[12];
    char * offset;
  } text;
  int x, y, height, width;
  canvas c;
  void * spra;
  char sprn[12];
  char fore, back, grey, sprwidth;
  list_tree * t;
} list_main;

#define LIST_SELECTED (1<<8)
#define LIST_FADED (1<<9)
#define LIST_OPENED (1<<10)
#define LIST_INDIRECT (1<<11)
#define LIST_OPENABLE (1<<7)

#define LIST_NOITEM (-1)

extern char listspr_minus[];
extern char listspr_plus[];

list_tree * list_foundlist;
int list_founditem;

static void __reverse_count(list_tree * t, int level, int flags)
{
  register int i;

  for(i = t -> current - 1; i >= 0; i --) {
    if((t -> item[i].t) && (t -> item[i].flags & LIST_OPENED)) __reverse_count(t -> item[i].t, level + 1, flags);
    flags |= (1 << level);
    t -> item[i].dispflags = flags;
  }
}

static void __update_tree(list_tree * t, int * h, int x, int * y)
{
  register int i, hh;

  hh = * h;
  t -> x = x;
  t -> y = * y;
  for(i = 0; i < t -> current; i ++) {
    t -> item[i].y = * y;
    * y -= 40;
    * h += 40;
    if(t -> item[i].flags & LIST_OPENED) {
      __update_tree((list_tree *)t -> item[i].t, h, x + 40, y);
    }
  }
  t -> height = * h - hh;
}

static void __update_parent(list_main * m)
{
  int h = 40, y = m -> y - 40;
  if(m -> flags & LIST_OPENED) __update_tree(m -> t, & h, m -> x + 40, & y);
  m -> height = h;
}

static void __spread_parent(list_tree * t, list_main * m)
{
  register int i;

  for(i = 0; i < t -> current; i ++) {
    if(t -> item[i].t) {
      __spread_parent((list_tree *)t -> item[i].t, m);
    }
  }
  t -> parent = m;
}

list_tree * list_create(int items, int flags)
{
  int size = sizeof(list_tree) + (items - 1) * sizeof(list_item);
  list_tree * t;

  t = rmalloc(size);
  if(t) {
    memset(t, 0, size);
    t -> items = items;
  }
  return t;
}

void list_destroy(list_tree * m)
{
  rfree(m);
}

int list_height(list_main * m)
{
  return m -> height;
}

void list_add(list_tree * t, char * txt, void * spra, char * sprn, void * handle)
{
  int i = t -> current;

  if(main_strlen(txt) > 11) {
    t -> item[i].text.offset = txt;
    t -> item[i].flags = LIST_INDIRECT;
  } else {
    main_strcpy(t -> item[i].text.name, txt);
    t -> item[i].flags = 0;
  }
  t -> item[i].spra = spra;
  main_strcpy(t -> item[i].sprn, sprn);
  t -> item[i].t = 0;
  t -> item[i].handle = handle;
  t -> current ++ ;
  t -> height += 48;
  if(t -> parent) {
    list_main * m = t -> parent;
    __update_parent(m);
    __reverse_count(m -> t, 0, 0);
  }
}

void list_changesprite(list_tree * t, int item, void * spra, char * sprn)
{
  list_main * m = t -> parent;

  t -> item[item].spra = spra;
  main_strcpy(t -> item[item].sprn, sprn);
  canvas_redraw(m -> c, m -> x - 4, t -> item[item].y - 44, m -> x + m -> width + 4, t -> item[item].y + 44);
}

list_main * list_attachtocanvas(list_tree * t, canvas c, char * name, void * spra, char * sprn, int fore, int back, int grey, int x, int y)
{
  list_main * m = rmalloc(sizeof(list_main));
  if(m == 0) return 0;

  m -> t = t;
  m -> c = c;
  m -> spra = spra;
  memcpy(m -> sprn, sprn, 12);
  if(main_strlen(name) < 12) {
    main_strcpy(m -> text.name, name);
    m -> flags = 0;
  } else {
    m -> text.offset = name;
    m -> flags = LIST_INDIRECT;
  }
  m -> x = x;
  m -> y = y;
  m -> height = 0;
  m -> fore = fore;
  m -> back = back;
  m -> grey = grey;
  m -> sprwidth = 38;
  m -> width = 512;
  m -> flags = LIST_OPENED | LIST_OPENABLE;
  __spread_parent(m -> t, m);
  __update_parent(m);
  __reverse_count(m -> t, 0, 0);
  return m;
}

void list_setwidth(list_main * m, int w)
{
  m -> width = w;
}

void list_setspritewidth(list_main * m, int w)
{
  m -> sprwidth = w;
}

void list_attachlist(list_tree * p, int pos, list_tree * c)
{
  register list_main * m;

  p -> item[pos].t = c;
  p -> item[pos].flags |= LIST_OPENABLE;
  p -> item[pos].flags &= (- 1 ^ LIST_OPENED);

  if(c) {
    if(p -> parent) {
      m = c -> parent = p -> parent;
      __spread_parent(c, p -> parent);
      __update_parent(p -> parent);
      __reverse_count(m -> t, 0, 0);
    }
  }
}

static void __draw_item(char * text, void * spra, char * sprn, int x, int y, int flags, int dispflags, list_main * m, int last)
{
  int xo, mx;
  icon_data d;
  int self = 0;

  if(flags & LIST_SELECTED) self = 1 << 21;

  if(dispflags) {
    mx = -1;
    if(last) {
      for(xo = 0; xo < 30; xo ++) {
	if(dispflags & (1 << xo)) {
	  mx = xo;
	}
      }
    }
    draw_gcol(m -> fore);
    for(xo = 0; xo < 30; xo ++) {
      if(dispflags & (1 << xo)) {
	draw_move(m -> x + (m -> sprwidth / 2) + xo * (m -> sprwidth), y);
	if(xo != mx) {
	  draw_lineto(0, - 40);
	} else {
	  draw_lineto(0, - 20);
	}
      }
    }
    draw_move(x/* + m -> sprwidth / 2 */, y - 20);
    draw_lineto(- m -> sprwidth / 2, 0);
  }
  if((flags & LIST_SELECTED) && ((flags & LIST_FADED) == 0)) {
    draw_gcol(m -> fore);
    draw_move(x - 8, y - 40);
    draw_plot(0x65, m -> x + m -> width, y - 1);
  }
  if(spra) {
    d.indirecttext.text = sprn;
    d.indirecttext.valid = (char *)spra;
    d.indirecttext.length = 1;
    if(flags & LIST_FADED) {
      draw_ploticon(x, y - 38, x + m -> sprwidth, y + 40, 0x1700010a | (1 << 22), & d);
    } else {
      draw_ploticon(x, y - 38, x + m -> sprwidth, y + 40, 0x1700010a | self, & d);
    }
    xo = m -> sprwidth;
    if(flags & LIST_OPENABLE) {
      if(flags & LIST_OPENED) {
	d.indirecttext.text = listspr_minus + 16;
	d.indirecttext.valid = listspr_minus;
      } else {
	d.indirecttext.text = listspr_plus + 16;
	d.indirecttext.valid = listspr_plus;
      }
      d.indirecttext.length = 0;
      if(flags & LIST_FADED) {
	draw_ploticon(x, y - 38, x + m -> sprwidth, y + 40, 0x17000102 | (1 << 22), & d);
      } else {
	draw_ploticon(x, y - 38, x + m -> sprwidth, y + 40, 0x17000102 | self, & d);
      }
    }  } else {
      xo = 0;
    }
  d.indirecttext.text = text;
  d.indirecttext.valid = (char *) - 1;
  d.indirecttext.length = 12;
  if(flags & LIST_FADED) {
    draw_ploticon(x + xo, y - 36, x + xo + 2047, y, (m -> grey << 24) | (m -> back << 28) | 0x111, & d);
  } else {
    if(flags & LIST_SELECTED) {
      draw_ploticon(x + xo, y - 36, x + xo + main_strwidth(text) + 16, y, (m -> back << 24) | (m -> fore << 28) | 0x131, & d);
    } else {
      draw_ploticon(x + xo, y - 36, x + xo + 2047, y, (m -> fore << 24) | (m -> back << 28) | 0x111, & d);
    }
  }
}

static void __draw_list(list_tree * t, int x, list_main * m, int y0, int y1, int greyf)
{
  register int i;
  static int _fore, _back, j;

#define ti t->item[i]

  for(i = 0; i < t -> current; i ++) {
    if((y0 < ti.y + 4) && (y1 > ti.y - 44)) {
      if(ti.flags & LIST_INDIRECT) {
	__draw_item(ti.text.offset, ti.spra, ti.sprn, x, ti.y, ti.flags | greyf, ti.dispflags, m, i == t -> current - 1);
      } else {
	__draw_item(ti.text.name, ti.spra, ti.sprn, x, ti.y, ti.flags | greyf, ti.dispflags, m, i == t -> current - 1);
      }
      if(t -> p) {
	j = 0;
	if((ti.flags & LIST_FADED) || (greyf)) {
	  j |= 1 << 22;
	  _fore = m -> grey;
	  _back = m -> back;
	} else {
	  if(ti.flags & LIST_SELECTED) {
	    j |= 1 << 21;
	    _fore = m -> back;
	    _back = m -> fore;
	  } else {
	    _fore = m -> fore;
	    _back = m -> back;
	  }
	}
	t -> p(t, i, x, ti.y, ti.handle, j, _fore, _back);
      }
    }
    if(ti.flags & LIST_OPENED) {
      if(ti.flags & LIST_FADED) {
	__draw_list(ti.t, x + m -> sprwidth, m, y0, y1, LIST_FADED);
      } else {
	__draw_list(ti.t, x + m -> sprwidth, m, y0, y1, greyf);
      }
    }
  }
#undef ti
}

void list_draw(list_main * m, int x0, int y0, int x1, int y1)
{
  int x, y;
  //icon_data d;

  if((x1 < m -> x) || (y0 > m -> y)) return;
  x = m -> x;
  y = m -> y;
  //d.indirecttext.length = 1;
  if((y0 < y + 4) && (y1 > y - 40)) {
    if(m -> flags & LIST_INDIRECT) {
      __draw_item(m -> text.offset, m -> spra, m -> sprn, m -> x, m -> y, m -> flags, 0, m, 0);
    } else {
      __draw_item(m -> text.name, m -> spra, m -> sprn, m -> x, m -> y, m -> flags, 0, m, 0);
    }
  }

  if(m -> flags & LIST_OPENED) __draw_list(m -> t, x + m -> sprwidth, m, y0, y1, m -> flags & LIST_FADED);
}

void list_fade(list_tree * t, int i, int flag)
{
  int flags = t -> item[i].flags;
  list_main * m = t -> parent;
  list_tree * t2;

  if(flag) {
    t -> item[i].flags |= LIST_FADED;
  } else {
    t -> item[i].flags &= (LIST_FADED ^ - 1);
  }
  if((flags != t -> item[i].flags) && (m)) {
    if(t -> item[i].flags & LIST_OPENED) {
      t2 = t -> item[i].t;
      canvas_redraw(m -> c, m -> x - 4, t -> item[i].y - 44 - t2 -> height, m -> x + m -> width + 4, t -> item[i].y + 44);
    } else {
      canvas_redraw(m -> c, m -> x - 4, t -> item[i].y - 44, m -> x + m -> width + 4, t -> item[i].y + 44);
    }
  }
}

void list_select(list_tree * t, int i, int flag)
{
  int flags = t -> item[i].flags;
  list_main * m = t -> parent;

  if(flag) {
    t -> item[i].flags |= LIST_SELECTED;
  } else {
    t -> item[i].flags &= (LIST_SELECTED ^ - 1);
  }
  if((flags != t -> item[i].flags) && (m)) {
    canvas_redraw(m -> c, m -> x - 4, t -> item[i].y - 44, m -> x + m -> width + 4, t -> item[i].y + 44);
  }
}

void list_redrawitem(list_tree * t, int i)
{
  list_main * m = t -> parent;
  canvas_redraw(m -> c, m -> x - 4, t -> item[i].y - 44, m -> x + m -> width + 4, t -> item[i].y + 44);
}

void list_open(list_tree * t, int i, int flag)
{
  int flags = t -> item[i].flags;
  list_main * m = t -> parent;
  int h1, h2;

  if(flag) {
    t -> item[i].flags |= LIST_OPENED;
  } else {
    t -> item[i].flags &= (LIST_OPENED ^ - 1);
  }
  if((flags != t -> item[i].flags) && (m)) {
    h1 = m -> height;
    __update_parent(m);
    __reverse_count(m -> t, 0, 0);
    h2 = m -> height;
    if(h1 < h2) {
      canvas_redraw(m -> c, m -> x - 4, m -> y - h2 - 44, m -> x + m -> width + 4, t -> item[i].y + 44);
    } else {
      canvas_redraw(m -> c, m -> x - 4, m -> y - h1 - 44, m -> x + m -> width + 4, t -> item[i].y + 44);
    }
  }
}

int list_selected(list_tree * t, int i)
{
  return(t -> item[i].flags & LIST_SELECTED) == LIST_SELECTED;
}

int list_faded(list_tree * t, int i)
{
  return(t -> item[i].flags & LIST_FADED) == LIST_FADED;
}

int list_opened(list_tree * t, int i)
{
  return(t -> item[i].flags & LIST_OPENED) == LIST_OPENED;
}

void * list_handle(list_tree * t, int i)
{
  return t -> item[i].handle;
}

static void __selected_items(list_tree * t, int * count)
{
  register int i;
  for(i = 0; i < t -> current; i ++) {
    if(t -> item[i].flags & LIST_SELECTED) (* count) ++ ;
    if((t -> item[i].flags & LIST_OPENABLE) && (t -> item[i].t)) __selected_items(t -> item[i].t, count);
  }
}

int list_selecteditems(list_main * m)
{
  int count = 0;

  if(m -> t) {
    __selected_items(m -> t, & count);
  }
  return count;
}

void list_clear(list_tree * t)
{
  t -> current = 0;
}

void list_reattach(list_main * m, list_tree * t)
{
  int h1, h2;
  h1 = m -> height;
  m -> t = t;
  __spread_parent(t, m);
  __update_parent(m);
  __reverse_count(m -> t, 0, 0);
  h2 = m -> height;
  if(h1 < h2) {
    canvas_redraw(m -> c, m -> x - 4, m -> y - h2 - 44, m -> x + m -> width + 4, m -> y + 4);
  } else {
    canvas_redraw(m -> c, m -> x - 4, m -> y - h1 - 44, m -> x + m -> width + 4, m -> y + 4);
  }
}

void list_setprintproc(list_tree * t, list_print p)
{
  t -> p = p;
  if(t -> parent) {
    list_main * m = t -> parent;
    canvas_redraw(m -> c, m -> x - 4, t -> y - t -> height - 4, m -> x + m -> width + 4, t -> y + 4);
  }
}

static list_tree * __findlist(list_tree * t, int y)
{
  register int i;
  register list_tree * q;

  for(i = 0; i < t -> current; i ++) {
    if((t -> item[i].y > y) && (t -> item[i].y <= y + 40)) {
      if(t -> item[i].flags & LIST_FADED) return 0;
      list_foundlist = t;
      list_founditem = i;
      return t;
    }
    if((t -> item[i].flags & LIST_OPENED) && ((t -> item[i].flags & LIST_FADED) == 0)) {
      q = __findlist(t -> item[i].t, y);
      if(q) return q;
    }
  }
  return 0;
}

list_tree * list_findlist(list_main * m, int y)
{
  list_foundlist = 0;
  list_founditem = -1;
  if(y > m -> y - 40) return 0; /* Above first */
  if(y < m -> y - m -> height) return 0; /* Below last */
  if((m -> flags & LIST_OPENED) == 0) return 0;
  return __findlist(m -> t, y);
}

int list_finditem(list_tree * t, int y)
{
  register int i;

  for(i = 0; i < t -> current; i ++) {
    if((t -> item[i].y > y) && (t -> item[i].y <= y + 40)) {
      if(t -> item[i].flags & LIST_FADED) return 0;
      return i;
    }
  }
  return LIST_NOITEM;
}

void list_deselectall(list_tree * t)
{
  int i;

  for(i = 0; i < t -> current; i ++) {
    if(t -> item[i].flags & LIST_SELECTED) {
      list_select(t, i, 0);
    }
    if(t -> item[i].t) list_deselectall(t -> item[i].t);
  }
}

int list_handleclick(list_main * m, canvas_mousestr * s)
{
  list_tree * t;
  int i;

  if((s -> y > m -> y) || (s -> x < m -> x)) return - 1;
  t = list_findlist(m, s -> y);
  if(t == 0) return - 1;
  i = list_founditem;
  if(i == LIST_NOITEM) return - 1;
  switch(s -> b) {
  case 256: /* Adjust pressed */
    t -> item[i].flags ^= LIST_SELECTED;
    canvas_redraw(m -> c, m -> x - 4, t -> item[i].y - 44, m -> x + m -> width + 4, t -> item[i].y + 4);
    if(t -> item[i].flags & LIST_SELECTED) {
      return 4;
    } else {
      return 5;
    }
    break;
  case 1024: /* Select pressed (turn off the others) */
    if(m -> t) list_deselectall(m -> t);
    list_select(t, i, 1);
    return 4;
    break;
  case 4 : /* Double - click select (load) */
    return 3;
    break;
  case 1 : /* Double - click adjust (close / open) */
    if(t -> item[i].flags & LIST_OPENABLE) {
      if(t -> item[i].t) {
	list_open(t, i, (t -> item[i].flags & LIST_OPENED) == 0);
	if(list_opened(t, i)) return 1; else return 2;
      } else {
	return 6;
      }
    }
    return 3;
    break;
  }
  return 0;
}

int list_findselected(list_tree * t, int i)
{
  int c;

  for(c = i + 1; c < t -> current; c ++) {
    if(t -> item[c].flags & LIST_SELECTED) {
      return c;
    }
  }
  return - 1;
}

int list_items(list_tree * t)
{
  return t -> current;
}

int list_ycoordinate(list_tree * t, int i)
{
  return t -> item[i].y - 44;
}
